/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package tabs;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Stroke;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;
import java.awt.image.AffineTransformOp;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.ResourceBundle;

import javax.imageio.ImageIO;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Scale;

import core.Loader;

import blending.BlendingComposite;

import utils.ImageUtils;
import utils.StockIcon;

/**
 * Instances of this class represents a tab that
 * can be used to paint the eyes shadow
 *
 */
public class ShadowTab extends TabContainer {

	/**
	 * Instances of this class represents a state of the {@link ShadowTab}
	 */
	private class ShadowHistoryStep {

		/**
		 * The left eye coordinates
		 */
		private final Rectangle leftEyeCoord;

		/**
		 * The right eye coordinates
		 */
		private final Rectangle rightEyeCoord;

		/**
		 * The selected button
		 */
		private final ColorButton selectedColorButton;

		/**
		 * The opacity value of the shadow effect
		 */
		private final int shadowOpacityScaleValue;

		/**
		 * The shadow image size
		 */
		private final int shadowSizeScaleValue;

		/**
		 * Creates a new instance of this class to save the
		 * {@link ShadowTab} given state
		 * @param leftEyeCoord the left eye coordinates
		 * @param rightEyeCoord the right eye coordinates
		 * @param colorButton the selected button
		 * @param shadowOpacityScaleValue the opacity value of the shadow effect
		 * @param shadowSizeScaleValue the shadow image size
		 */
		public ShadowHistoryStep(Rectangle leftEyeCoord, Rectangle rightEyeCoord, ColorButton colorButton, int shadowOpacityScaleValue, int shadowSizeScaleValue) {

			this.leftEyeCoord = new Rectangle(leftEyeCoord);
			this.rightEyeCoord = new Rectangle(rightEyeCoord);

			//search for the reference of the actual button instead of the reference of the selectedButton
			ColorButton selectedColorButton = null;
			if (colorButton != null) {
				for (Control c : colorButton.getParent().getChildren()) {
					if (c instanceof ColorButton) {
						ColorButton cb = (ColorButton) c;
						if (cb.equals(colorButton)) {
							selectedColorButton = cb;
							break;
						}
					}
				}
			}
			this.selectedColorButton = selectedColorButton;
			this.shadowOpacityScaleValue = shadowOpacityScaleValue;
			this.shadowSizeScaleValue = shadowSizeScaleValue;
		}

	}

	/**
	 * The opacity that will be used to draw the shadow
	 */
	private static float ALPHA_COMPOSITE_OPACITY = 0.4f;

	/**
	 * Alpha composite used to draw the shadow images
	 */
	private AlphaComposite currentCompositeOpacity;

	/**
	 * The opacity of the shadow images bounding box
	 */
	private static float BOUNDS_OPACITY = 0.2f;

	/**
	 * Alpha composite used to draw the shadow images bounding box
	 */
	private AlphaComposite boundsCompositeOpacity;

	/**
	 * messages that are to be used by the entire GUI
	 */
	private ResourceBundle messages;

	/**
	 * A list with all the colors used by ColorButton
	 * For each color in this list one ColorButton will be created
	 */
	private Color[] colorsList = {
			new Color(Display.getCurrent(), 236, 229, 236),
			new Color(Display.getCurrent(), 206, 233, 240),
			new Color(Display.getCurrent(), 147, 219, 237),
			new Color(Display.getCurrent(), 80, 192, 220),
			new Color(Display.getCurrent(), 53, 154, 179),
			new Color(Display.getCurrent(), 23, 109, 130),
			new Color(Display.getCurrent(), 229, 201, 187),
			new Color(Display.getCurrent(), 203, 169, 152),
			new Color(Display.getCurrent(), 189, 137, 111),
			new Color(Display.getCurrent(), 166, 104, 73),
			new Color(Display.getCurrent(), 140, 79, 49),
			new Color(Display.getCurrent(), 114, 50, 18),
			new Color(Display.getCurrent(), 250, 193, 210),
			new Color(Display.getCurrent(), 231, 169, 184),
			new Color(Display.getCurrent(), 219, 153, 167),
			new Color(Display.getCurrent(), 196, 119, 135),
			new Color(Display.getCurrent(), 191, 71, 96),
			new Color(Display.getCurrent(), 159, 67, 92),
			new Color(Display.getCurrent(), 238, 205, 226),
			new Color(Display.getCurrent(), 217, 182, 214),
			new Color(Display.getCurrent(), 196, 144, 192),
			new Color(Display.getCurrent(), 193, 103, 186),
			new Color(Display.getCurrent(), 169, 70, 161),
			new Color(Display.getCurrent(), 129, 79, 104),
			new Color(Display.getCurrent(), 247, 185, 196),
			new Color(Display.getCurrent(), 239, 152, 152),
			new Color(Display.getCurrent(), 229, 123, 123),
			new Color(Display.getCurrent(), 207, 82, 94),
			new Color(Display.getCurrent(), 196, 55, 68),
			new Color(Display.getCurrent(), 167, 32, 46),
			new Color(Display.getCurrent(), 255, 190, 172),
			new Color(Display.getCurrent(), 239, 153, 130),
			new Color(Display.getCurrent(), 228, 120, 91),
			new Color(Display.getCurrent(), 213, 102, 73),
			new Color(Display.getCurrent(), 201, 96, 68),
			new Color(Display.getCurrent(), 179, 87, 62),
			new Color(Display.getCurrent(), 241, 209, 158),
			new Color(Display.getCurrent(), 231, 196, 141),
			new Color(Display.getCurrent(), 236, 185, 104),
			new Color(Display.getCurrent(), 234, 160, 44),
			new Color(Display.getCurrent(), 222, 145, 24),
			new Color(Display.getCurrent(), 200, 129, 19),
			new Color(Display.getCurrent(), 194, 241, 159),
			new Color(Display.getCurrent(), 160, 224, 115),
			new Color(Display.getCurrent(), 109, 181, 56),
			new Color(Display.getCurrent(), 62, 124, 17),
			new Color(Display.getCurrent(), 41, 90, 5),
			new Color(Display.getCurrent(), 47, 66, 37),
			new Color(Display.getCurrent(), 212, 212, 212),
			new Color(Display.getCurrent(), 184, 184, 184),
			new Color(Display.getCurrent(), 159, 157, 158),
			new Color(Display.getCurrent(), 114, 114, 114),
			new Color(Display.getCurrent(), 77, 77, 77),
			new Color(Display.getCurrent(), 0, 0, 0)
	};

	//the left and right eye shadow images
	private BufferedImage shadowLeftEye = null;
	private BufferedImage shadowRightEye = null;

	//the left and right coordinates of the shadow images
	private Rectangle leftEyeCoord = null;
	private Rectangle rightEyeCoord = null;

	/**
	 * Used to set the upper and color buttons composite
	 */
	private TabGuiUtility tabGui;

	/**
	 * the tab's main composite
	 */
	private Composite mainComposite;

	/**
	 * All the colored buttons in use by this tab
	 */
	private ArrayList<ColorButton> coloredButtonList;

	/**
	 * The shadow original image file read from the file system
	 */
	private BufferedImage shadowFile;

	//mouse listeners in the drawing area
	private MouseListener mouseListener;
	private MouseMotionListener mouseMotionListener;

	//flags to indicate if the mouse is over the left or right eye shadow
	protected boolean mouseOverLeftShadow;
	protected boolean mouseOverRightShadow;

	//used to help with the dragging of the shadow images over the drawing area
	protected int startX;
	protected int startY;
	protected Point oldLeftEyePoint;
	protected Point oldRightEyePoint;

	/**
	 * Composite for the bottom controls
	 */
	private Composite bottomComposite;

	//bottom composite label separators
	private Label labelSeparatotUp;
	private Label labelSeparatorDown;

	//Scale for the size and opacity of the eye shadow images
	private Scale sizeScale;
	private Scale opacityScale;

	//labels for the size and opacity scale
	private Label labelSizeScale;
	private Label labelOpacityScale;

	/**
	 * Button for clearing this tab's contents on the drawing area
	 */
	private Button clearButton;

	/**
	 * The eyes image created using the given eyes paths
	 */
	private BufferedImage eyesImage;

	/**
	 * flag to indicate whether the tab has or not access to call for redrawing the image area
	 */
	protected boolean imageAreaRedrawAllowed;

	/**
	 * A reference to this instance class
	 */
	private ShadowTab thisShadowTab = this;

	/**
	 * Constructs a new instance of this class given its parent,
	 * a style value describing its behavior and appearance and
	 * this instance's position in the {@link TabsContainer} array
	 *
	 * @param parent a {@link TabsContainer} which will be the parent of the new instance
	 * @param style the style of the tab
	 * @param tabLayerPosition the tab's position in the {@link TabsContainer} array
	 *
	 */
	public ShadowTab(TabsContainer parent, int style, int tabLayerPosition) {
		super(parent, style, tabLayerPosition);

		messages = Loader.getMessages();

		//the 2 things needed to be done by every implementation of TabComposite
		super.setTabImage(new Image(Display.getCurrent(), System.getProperty("program.pixmapspath") + "tab_icon_shadow.png"));
		super.setTabToolTipText(messages.getString("SHADOW_TAB_TOOLTIP"));

		addDisposeListener(new DisposeListener() {
			public void widgetDisposed(DisposeEvent e) {
				for (Color c : colorsList) {
					c.dispose();
				}
			}
		});

		imageAreaRedrawAllowed = true;

		//create the tab's GUI
		tabGui = new TabGuiUtility(this);
		mainComposite = this;
		tabGui.createUpperComposite(super.getTabToolTipText());
		coloredButtonList = tabGui.createColorButtonsComposite(colorsList);
		tabGui.setSelectedColorButton(null);
		createBottomComposite();

		//default opacity
		currentCompositeOpacity = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, ALPHA_COMPOSITE_OPACITY);
		boundsCompositeOpacity = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, BOUNDS_OPACITY);

		//read the shadow image file
		createColoredShadow();

		extendColorButtonsBehavior();
		leftEyeCoord = new Rectangle();
		rightEyeCoord = new Rectangle();
	}

	/**
	 * This method initializes bottomComposite	
	 */
	public void createBottomComposite() {

		//grid data for the clear button
		GridData gridDataClearButton = new GridData();
		gridDataClearButton.grabExcessHorizontalSpace = true;
		gridDataClearButton.horizontalSpan = 2;
		gridDataClearButton.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the separator label up
		GridData gridDataSeparatorLabelUp = new GridData();
		gridDataSeparatorLabelUp.grabExcessHorizontalSpace = true;
		gridDataSeparatorLabelUp.horizontalSpan = 2;
		gridDataSeparatorLabelUp.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the separator label down
		GridData gridDataSeparatorLabelDown = new GridData();
		gridDataSeparatorLabelDown.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataSeparatorLabelDown.horizontalSpan = 2;
		gridDataSeparatorLabelDown.grabExcessHorizontalSpace = true;

		//grid data for the opacity scale
		GridData gridDataOpacityScale = new GridData();
		gridDataOpacityScale.grabExcessHorizontalSpace = true;
		gridDataOpacityScale.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the size scale
		GridData gridDataSizeScale = new GridData();
		gridDataSizeScale.grabExcessHorizontalSpace = true;
		gridDataSizeScale.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.verticalAlignment = org.eclipse.swt.layout.GridData.END;
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataComposite.grabExcessVerticalSpace = true;

		//grid layout for the composite
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.numColumns = 2;

		//initialize the composite
		bottomComposite = new Composite(mainComposite, SWT.NONE);
		bottomComposite.setLayout(gridLayoutComposite);
		bottomComposite.setLayoutData(gridDataComposite);

		//initialize the upper label separator
		labelSeparatotUp = new Label(bottomComposite, SWT.SEPARATOR | SWT.HORIZONTAL);
		labelSeparatotUp.setLayoutData(gridDataSeparatorLabelUp);

		//initialize the clear button
		clearButton = new Button(bottomComposite, SWT.NONE);
		clearButton.setText(messages.getString("SHADOW_TAB_CLEAR"));
		clearButton.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_CLEAR, StockIcon.SIZE_MENU, true)));
		clearButton.setLayoutData(gridDataClearButton);

		//initialize the size scale label
		labelSizeScale = new Label(bottomComposite, SWT.NONE);
		labelSizeScale.setText(messages.getString("SHADOW_TAB_SIZE"));

		//initialize the size scale
		sizeScale = new Scale(bottomComposite, SWT.NONE);
		sizeScale.setSelection(50);
		sizeScale.setMinimum(20);
		sizeScale.setMaximum(80);
		sizeScale.setLayoutData(gridDataSizeScale);

		//initialize the opacity scale label
		labelOpacityScale = new Label(bottomComposite, SWT.NONE);
		labelOpacityScale.setText(messages.getString("SHADOW_TAB_OPACITY"));

		//initialize the opacity scale
		opacityScale = new Scale(bottomComposite, SWT.NONE);
		opacityScale.setSelection(50);
		opacityScale.setLayoutData(gridDataOpacityScale);

		//initialize the lower separator label
		labelSeparatorDown = new Label(bottomComposite, SWT.SEPARATOR | SWT.HORIZONTAL);
		labelSeparatorDown.setLayoutData(gridDataSeparatorLabelDown);

		//clear button listener
		clearButton.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				//TODO check if history step is created only when needed
				if (e.widget instanceof Button && shadowLeftEye != null && shadowRightEye != null) {
					tabsContainer.createHistoryStep();
					clearImageArea();
					if (imageAreaRedrawAllowed) {
						tabsContainer.redrawImageArea(thisShadowTab);
					}
				}
			}
		});

		//opacity scale listeners
		opacityScale.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				double opacityInterval = 0.3;
				double adjustInterval = (((float)opacityScale.getSelection() / opacityScale.getMaximum()) * opacityInterval) - (opacityInterval / 2.0);
				currentCompositeOpacity = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (float)(ALPHA_COMPOSITE_OPACITY + adjustInterval));
				if (imageAreaRedrawAllowed) {
					tabsContainer.redrawImageArea(thisShadowTab);
				}
			}
		});
		opacityScale.addListener(SWT.MouseDown, new Listener() {
			public void handleEvent(Event event) {
				tabsContainer.createHistoryStep();
			}
		});

		//size scale listeners
		sizeScale.addSelectionListener(new org.eclipse.swt.events.SelectionAdapter() {
			public void widgetSelected(org.eclipse.swt.events.SelectionEvent e) {
				createColoredShadows();
				if (imageAreaRedrawAllowed) {
					tabsContainer.redrawImageArea(thisShadowTab);
				}
			}
		});
		sizeScale.addListener(SWT.MouseDown, new Listener() {
			public void handleEvent(Event event) {
				tabsContainer.createHistoryStep();
			}
		});
	}

	/**
	 * Color buttons should do more than just updating the {@link ColorDifference}
	 * All color buttons listeners are implemented here
	 */
	private void extendColorButtonsBehavior() {
		for (final ColorButton button : coloredButtonList) {
			button.addMouseListener(new org.eclipse.swt.events.MouseAdapter() {

				public void mouseDown(org.eclipse.swt.events.MouseEvent e) {
					//TODO history is created but user may click cancel on EyesTracer and nothing is drawn
					tabsContainer.createHistoryStep();
					//trick the flags so that borders are shown when changing shadow colors
					mouseOverLeftShadow = true;
					mouseOverRightShadow = true;
					if (imageAreaRedrawAllowed) {
						tabsContainer.redrawImageArea(thisShadowTab);
					}
				}

				public void mouseUp(org.eclipse.swt.events.MouseEvent e) {
					if (!hasEyesPath()) {
						tabsContainer.setShadowPath();
					}
					try {
						Thread.sleep(100);
						createColoredShadows();
						if (imageAreaRedrawAllowed) {
							tabsContainer.redrawImageArea(thisShadowTab);
						}
						Thread.sleep(100);
					} catch (InterruptedException e1) {
						// TODO Auto-generated catch block
						e1.printStackTrace();
					}
					//restore the flags values
					mouseOverLeftShadow = false;
					mouseOverRightShadow = false;
					createColoredShadows();
					if (imageAreaRedrawAllowed) {
						tabsContainer.redrawImageArea(thisShadowTab);
					}
				}
			});
		}
	}

	/**
	 * Reads the shadow image from the file system
	 * resizes it and blends it with the current selected color
	 * from the colored button
	 * @return true if image creation succeeded
	 */
	private boolean createColoredShadow() {

		double widthFactor = sizeScale.getSelection() / 50.0;
		double heightFactor = sizeScale.getSelection() / 50.0;
		AffineTransform tx = new AffineTransform(); //use affine transform class to do the resizing
		tx.scale(widthFactor, heightFactor);
		AffineTransformOp op = new AffineTransformOp(tx, AffineTransformOp.TYPE_BICUBIC);

		try {
			shadowFile = ImageIO.read(new File(System.getProperty("program.pixmapspath") + "tab_shadow_H3.png"));
			//TODO resize should keep the shadow in the middle and not to extend to lower right
			shadowFile = op.filter(shadowFile, null);
			if (leftEyeCoord != null && rightEyeCoord != null) {
				leftEyeCoord.width = shadowFile.getWidth();
				leftEyeCoord.height = shadowFile.getHeight();
				rightEyeCoord.width = shadowFile.getWidth();
				rightEyeCoord.height = shadowFile.getHeight();				
			}
			//appy the color to the shadow image
			if (tabGui.getCurrentColor() == null) return false;
			java.awt.Composite compositeToUse = Loader.getCompositeFor(BlendingComposite.COLOR_OVERLAY);
			shadowFile = ImageUtils.applyBlending(shadowFile, compositeToUse, tabGui.getCurrentColor().getRed(), tabGui.getCurrentColor().getGreen(), tabGui.getCurrentColor().getBlue());

		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
			return false;
		}
		return true;
	}

	/**
	 * Clears this tab's image in the drawing area
	 * and resets the needed variables
	 */
	private void clearImageArea() {
		tabGui.setSelectedColorButton(null);
		shadowLeftEye = null;
		shadowRightEye = null;
	}

	/**
	 * Set the eyes paths to the given one; if the new
	 * eyes paths is <code>null</code> nothing is done
	 * @param eyesPaths the new eyes paths or <code>null</code>
	 */
	public void setEyesPath(GeneralPath[] eyesPaths) {
		if (eyesPaths == null || eyesPaths[0] == null || eyesPaths[1] == null) return ;

		//create the left eye shadow coordinate rectangle
		leftEyeCoord = new Rectangle(
				eyesPaths[0].getBounds().x + eyesPaths[0].getBounds().width + 10 - shadowFile.getWidth(),
				eyesPaths[0].getBounds().y + eyesPaths[0].getBounds().height + 8 - shadowFile.getHeight(),
				shadowFile.getWidth(),
				shadowFile.getHeight()
		);

		//create the right eye shadow coordinate rectangle
		rightEyeCoord = new Rectangle(
				eyesPaths[1].getBounds().x - 10,
				eyesPaths[1].getBounds().y + eyesPaths[1].getBounds().height + 8 - shadowFile.getHeight(),
				shadowFile.getWidth(),
				shadowFile.getHeight()
		);

		//create a general path for clipping
		GeneralPath eyesGeneralPath = new GeneralPath(eyesPaths[0]);
		eyesGeneralPath.append(eyesPaths[1], false);

		eyesImage = ImageUtils.createSoftClipping(tabsContainer.getOriginalImage(), eyesGeneralPath);
		createColoredShadows();
		if (imageAreaRedrawAllowed) {
			tabsContainer.redrawImageArea(thisShadowTab);
		}
	}

	/**
	 * Check whether this tab instance has an associated eyes shadow path
	 * @return true if the eyes shadow path is found, false otherwise
	 */
	public boolean hasEyesPath() {
		return eyesImage != null;
	}

	/**
	 * Create the two shadows images from the shadow image file
	 */
	private void createColoredShadows() {
		if (eyesImage != null) {

			if (!createColoredShadow()) return ;

			//create the right eye shadow image
			shadowRightEye = new BufferedImage(shadowFile.getWidth(), shadowFile.getHeight(), BufferedImage.TYPE_INT_ARGB);
			Graphics2D g2dRightEye = shadowRightEye.createGraphics();
			g2dRightEye.drawImage(shadowFile, 0, 0, null);
			g2dRightEye.dispose();

			//flip the shadow file
			AffineTransform tx = AffineTransform.getScaleInstance(-1, 1);
			tx.translate(-shadowFile.getWidth(null), 0);
			AffineTransformOp op = new AffineTransformOp(tx, AffineTransformOp.TYPE_NEAREST_NEIGHBOR);
			shadowFile = op.filter(shadowFile, null);

			//create the left eye shadow image
			shadowLeftEye = new BufferedImage(shadowFile.getWidth(), shadowFile.getHeight(), BufferedImage.TYPE_INT_ARGB);
			Graphics2D g2dLeftEye = shadowLeftEye.createGraphics();
			g2dLeftEye.drawImage(shadowFile, 0, 0, null);
			g2dLeftEye.dispose();
		}
	}

	@Override
	public BufferedImage getLayerImage() {
		//create a copy of this tab's image and return it
		BufferedImage transparentShadowImage = null;
		if (shadowLeftEye != null && shadowRightEye != null) {
			transparentShadowImage = new BufferedImage(tabsContainer.getOriginalImage().getWidth(), tabsContainer.getOriginalImage().getHeight(), BufferedImage.TYPE_INT_ARGB);
			Graphics2D g2d = transparentShadowImage.createGraphics();
			drawShadowImage(g2d);
			g2d.dispose();
		}
		return transparentShadowImage;
	}

	/**
	 * Draws the shadow image using the given graphics context
	 * @param g2d the graphics context used for drawing operations
	 */
	private void drawShadowImage(Graphics2D g2d) {

		java.awt.Composite defaultComposite = g2d.getComposite();

		//set shadow image opacity
		g2d.setComposite(currentCompositeOpacity);

		//draw the left and right shadows
		g2d.drawImage(shadowLeftEye, leftEyeCoord.x, leftEyeCoord.y, null);
		g2d.drawImage(shadowRightEye, rightEyeCoord.x, rightEyeCoord.y, null);

		//restore the g2d original composite
		g2d.setComposite(defaultComposite);

		//draw the eyes above the shadow
		g2d.drawImage(eyesImage, 0, 0, null);

		//check if mouse is in shadow area and highlight the area
		//TODO check if cursor is visible when drawing the rectangle because of full screen preview
		if (mouseOverLeftShadow) {
			Stroke defaultStroke = g2d.getStroke();
			g2d.setStroke(new BasicStroke(2,1,1));
			g2d.setComposite(boundsCompositeOpacity);
			g2d.drawRect(leftEyeCoord.x, leftEyeCoord.y, leftEyeCoord.width, leftEyeCoord.height);
			g2d.setStroke(defaultStroke);
			g2d.setComposite(defaultComposite);
		}
		if (mouseOverRightShadow) {
			Stroke defaultStroke = g2d.getStroke();
			g2d.setStroke(new BasicStroke(2,1,1));
			g2d.setComposite(boundsCompositeOpacity);
			g2d.drawRect(rightEyeCoord.x, rightEyeCoord.y, rightEyeCoord.width, rightEyeCoord.height);
			g2d.setStroke(defaultStroke);
			g2d.setComposite(defaultComposite);
		}
	}

	@Override
	public void drawCurrentLayer(Graphics2D g2d) {
		//if there exists left and right shadows draw them to the drawing area
		if (shadowLeftEye != null && shadowRightEye != null) {
			drawShadowImage(g2d);
		}
	}

	@Override
	public void clear() {
		clearImageArea();
	}

	@Override
	public Object getCurrentState() {
		return new ShadowHistoryStep(
				leftEyeCoord,
				rightEyeCoord,
				tabGui.getSelectedButton(),
				opacityScale.getSelection(),
				sizeScale.getSelection()
		);
	}

	@Override
	public void setCurrentState(Object o) {
		if (o instanceof ShadowHistoryStep) {
			ShadowHistoryStep hs = (ShadowHistoryStep) o;

			imageAreaRedrawAllowed = false;

			tabGui.setSelectedColorButton(hs.selectedColorButton);
			leftEyeCoord = hs.leftEyeCoord;
			rightEyeCoord = hs.rightEyeCoord;

			opacityScale.setSelection(hs.shadowOpacityScaleValue);
//			opacityScale.notifyListeners(eventType, event) //TODO this is how should work instead of the 3 lines below 
			double opacityInterval = 0.3;
			double adjustInterval = (((float)opacityScale.getSelection() / opacityScale.getMaximum()) * opacityInterval) - (opacityInterval / 2.0);
			currentCompositeOpacity = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, (float)(ALPHA_COMPOSITE_OPACITY + adjustInterval));

			sizeScale.setSelection(hs.shadowSizeScaleValue);
			createColoredShadows();

			if (hs.selectedColorButton == null) {
				shadowLeftEye = null;
				shadowRightEye = null;
			}
			imageAreaRedrawAllowed = true;
		}
	}

	public MouseListener mouseListener() {
		if (mouseListener == null) {
			mouseListener = new MouseListener() {

				//unimplemented methods
				public void mouseClicked(MouseEvent e) {}
				public void mouseExited(MouseEvent e) {}

				public void mouseEntered(MouseEvent e) {
					e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.DEFAULT_CURSOR));					
				}

				public void mousePressed(MouseEvent e) {
					if (mouseOverLeftShadow || mouseOverRightShadow) {

						//save the tab's current state before changing anything
						tabsContainer.getDisplay().syncExec(new Runnable() {
							public void run() {
								tabsContainer.createHistoryStep();
							}
						});

						//enable dragging for the left eye shadow
						if (mouseOverLeftShadow) {
							startX = e.getX();
							startY = e.getY();
							oldLeftEyePoint = new Point(leftEyeCoord.x, leftEyeCoord.y);
							e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.MOVE_CURSOR));
						}

						//enable dragging for the right eye shadow
						if (mouseOverRightShadow) {
							startX = e.getX();
							startY = e.getY();
							oldRightEyePoint = new Point(rightEyeCoord.x, rightEyeCoord.y);
							e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.MOVE_CURSOR));
						}
					}
				}

				public void mouseReleased(MouseEvent e) {
					if (mouseOverLeftShadow || mouseOverRightShadow) {
						e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.HAND_CURSOR));
					} else {
						e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.DEFAULT_CURSOR));
					}
				}
			};
		}
		return mouseListener;
	}

	public MouseMotionListener mouseMotionListener() {
		if (mouseMotionListener == null) {
			mouseMotionListener = new MouseMotionListener() {

				public void mouseDragged(MouseEvent e) {
					if (mouseOverLeftShadow) {
						leftEyeCoord.x = e.getX() - startX + oldLeftEyePoint.x;
						leftEyeCoord.y = e.getY() - startY + oldLeftEyePoint.y;
					}
					if (mouseOverRightShadow) {
						rightEyeCoord.x = e.getX() - startX + oldRightEyePoint.x;
						rightEyeCoord.y = e.getY() - startY + oldRightEyePoint.y;
					}
					if (mouseOverLeftShadow || mouseOverRightShadow) {
						if (imageAreaRedrawAllowed) {
							tabsContainer.redrawImageArea(thisShadowTab);
						}
					}
				}

				public void mouseMoved(MouseEvent e) {
					if (shadowLeftEye == null || shadowRightEye == null) return ;

					//check if mouse is over shadows
					boolean oldLeft = mouseOverLeftShadow;
					boolean oldRight = mouseOverRightShadow;
					mouseOverLeftShadow = leftEyeCoord.contains(e.getPoint());
					mouseOverRightShadow = rightEyeCoord.contains(e.getPoint());
					if (mouseOverLeftShadow != oldLeft || mouseOverRightShadow != oldRight) {
						if (imageAreaRedrawAllowed) {
							tabsContainer.redrawImageArea(thisShadowTab);
						}
					}
					if (mouseOverLeftShadow || mouseOverRightShadow) {
						e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.HAND_CURSOR));
					} else {
						e.getComponent().setCursor(new java.awt.Cursor(java.awt.Cursor.DEFAULT_CURSOR));
					}
				}
			};
		}
		return mouseMotionListener;
	}

	public void originalImageChanged() {
		eyesImage = null;		
	}

	public void disposeData() {
		eyesImage = null;
		shadowFile = null;
		shadowLeftEye = null;
		shadowRightEye = null;
	}
}